'''
..
  $Id$
  
  Basic python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''
import io
import logging
import os
import pytz
from datetime import datetime, timedelta
import stat
import time

log = logging.getLogger(__name__)

_micro_factors = (1000000,100000,10000,1000,100,10,1)

def getCurrentTimeMillis():
    """
    Equivalent to Java's System.getCurrentTimeMillis()
    
    :return: milliseconds passed since the epoch(1970-01-01 0:00 GMT)
    """
    return int(time.time() * 1000)

def parse_datetime(s,tzinfo=None):
    '''
    Parse an ISO 8601 timestamp with seconds or milliseconds an a optional timezone.
    
    :param s: An ISO 8601 string.
    :param tzinfo: An optional timezone to apply for timestamps with no UTC offset.
    '''
    if s is None:
        return None
    
    try:
        
        year = int(s[0:4])
        
        if s[4] != '-':
            raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], no hyphen after year."%s)

        month = int(s[5:7])
        
        if s[7] != '-':
            raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], no hyphen after year."%s)

        day = int(s[8:10])
        
        if len(s) == 10:
            hour = 0
            minute = 0
            second = 0
            microsecond = 0
            tz = None    
        
        else:
            
            if s[10] == 'T':
   
                hour = int (s[11:13])
            
                if s[13] != ':':
                    raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], no colon after hour."%s)
        
                minute = int (s[14:16])
            
                if s[16] != ':':
                    raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], no colon after minute."%s)

                second = int (s[17:19])

                idx = 19

                if len(s) == idx:
                    microsecond = 0
                else:
                
                    if s[idx] == '.':
                    
                        idx += 1
                        while idx < len(s) and s[idx].isdigit():
                            idx += 1
                    
                        n = idx - 20
                    
                        if n > 6:
                            raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], more than 6 fractional digits."%s)

                        if n == 0:
                            microsecond = 0
                        else:
                            microsecond = int(s[20:idx]) * _micro_factors[n]
                    else:
                        microsecond = 0
            else:
                idx = 10
                hour = 0
                minute = 0
                second = 0
                microsecond = 0

            if idx >= len(s):
                tz = None
                
            elif s[idx] == 'Z':
                tz = pytz.utc
                idx += 1
                
                if idx != len(s):
                    raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], garbage after UTC timezone indicator."%s)
                
            else:
                
                if s[idx] == '+':
                    sign = 1
                elif s[idx] == '-':
                    sign = -1
                else:
                    raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], invalid timezone designator."%s)

                idx += 1
                
                tzhour = int(s[idx:idx+2])
                
                idx += 2
                
                if idx < len(s):
                    if s[idx] == ':':
                        idx += 1

                    tzmin = int(s[idx:idx+2]) if idx < len(s) else 0
                    
                    idx += 2
                    
                else:
                    tzmin = 0

                if idx != len(s):
                    raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], garbage after timezone designator."%s)
                
                tz = pytz.FixedOffset(sign * (tzhour * 60 + tzmin))
        
        dt = datetime(year,month,day,hour,minute,second,microsecond,tz)
            
        if tzinfo is None or tz is not None:
            return dt
        else:
            #
            # believe it or not, this is the only reliable way to convert
            # a zone-agnostic datetime to a datetime with the correct offset.
            #
            # 1. tzinfo.utcoffset() takes an agnostic datetime.
            # 2. transform the timestamp to a timestamp in UTC with the correct millis since the epoch
            # 3. Finally, apply the timezone offset using datetime.astimezone().
            #
            try:
                offset = tzinfo.utcoffset(dt)
            except pytz.AmbiguousTimeError:
                offset = tzinfo.utcoffset(dt.replace(hour=0,minute=0,second=0,microsecond=0))
            dt2 = dt.replace(tzinfo=pytz.utc) - offset
            return dt2.astimezone(tzinfo)

    except ValueError:
        raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], invalid number in one of the parts."%s)
    
    except IndexError:
        raise SyntaxError("Unrecognized ISO 8601 timestamp [%s], string is too short."%s)


def datetime_to_millis(dt):
    '''
    Convert a timestamp to integer milliseconds.
    
    :param dt: A datetime instance or None.
    '''
    if dt is None:
        return None
    
    return round(dt.timestamp()*1000.0)


def format_datetime(dt):
    '''
    Format a datetime instance ISO-8601 format with milliseconds.
    
    :param dt: A datetime instance.
    '''
    
    if dt is None:
        return None
    
    millis = dt.microsecond // 1000

    utcoff = dt.utcoffset()
    
    if utcoff is not None:
        off = utcoff.total_seconds()
        
        if off > 0:
            return "%04d-%02d-%02dT%02d:%02d:%02d.%03d+%02d:%02d"%(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,millis,off//3600,(off//60)%60)
        elif off < 0:
            return "%04d-%02d-%02dT%02d:%02d:%02d.%03d-%02d:%02d"%(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,millis,(-off)//3600,((-off)//60)%60)
        else:
            return "%04d-%02d-%02dT%02d:%02d:%02d.%03dZ"%(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,millis)
    else:
        return "%04d-%02d-%02dT%02d:%02d:%02d.%03d"%(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,millis)
        
def to_datetime_timestamp_millis(timestamp,tz):
    '''
    Convert a millisecond timestamp to a datetime instance.
    
    :param timestamp: A timestamp in milliseconds since the epoch ``1970-01-01T00:00:00Z``.
    :param tz: An optional timezone to apply for generating a UTC offset.
    :return: A datetime instance.
    '''
    
    if timestamp is None:
        return None
    
    return datetime.fromtimestamp(timestamp/1000.0,tz)
    
def format_timestamp_millis(timestamp,tz):
    '''
    Format a datetime instance in ISO-8601 format with milliseconds.
    
    :param timestamp: A timestamp in milliseconds since the epoch ``1970-01-01T00:00:00Z``.
    :param tz: An optional timezone to apply for rendering a UTC offset. 
    '''
    return format_datetime(to_datetime_timestamp_millis(timestamp,tz))

def _parse_duration_number_part(s,idx):
    
    have_dot = False
    
    eidx = idx
    
    c=None
    
    while eidx < len(s):
        
        c = s[eidx]
        
        if c in "YMDHSW":
            break
        
        if c == '.':
            
            if have_dot:
                raise SyntaxError("ISO 8601 duration [%s] contains part with two dots in number."%s)
        
            have_dot = True
        
        elif c < '0' or c >'9':
            raise SyntaxError("ISO 8601 duration [%s] contains part with an invalid number."%s)
            
        eidx += 1

    if c is None:
        raise SyntaxError("ISO 8601 duration [%s] contains trailing part without an identifier."%s)        

    if have_dot:
        return eidx+1,c,float(s[idx:eidx])
    else:
        return eidx+1,c,int(s[idx:eidx])

def parse_duration(s):
    '''
    Parse `An ISO8601 duration string <https://en.wikipedia.org/wiki/ISO_8601#Durations>`_
    
    This function returns datateim.timedelta instances, which means, that months are interpreted
    as 30 days and years as 365 days. Real dat/tome deltas are out of scope of this function.
    
    :param s: A string like ``P3M`` for three months or ``P1DT3H`` for one day and three hours.
    :return: A timedelta value.
    '''
    
    if len(s) < 1 or s[0] != 'P':
        raise SyntaxError("ISO 8601 duration [%s] does not start with 'P'."%s)
    
    idx = 1
    
    # ctor args for timedelta
    kwargs = {}
    # parse date fraction into days
    days = 0
    
    while idx < len(s):
        
        if s[idx] == 'T':
            idx += 1
            break

        oidx = idx
        idx,pfx,v = _parse_duration_number_part(s,idx)
        
        # special case 'PWn'
        if oidx == 1 and pfx == 'W':
            
            if idx != len(s):
                raise SyntaxError("ISO 8601 duration [%s] contains garbage after weeks designator."%s)
            
            return timedelta(days=v*7)
        
        if pfx == 'Y':
            days += v * 365
            
        elif pfx == 'M':
            days += v * 30
            
        elif pfx == 'D':
            days += v
        
        else:
            if idx != len(s):
                raise SyntaxError("ISO 8601 duration [%s] contains invalid date part [%s]."%(s,pfx))
    
    if days != 0:
        kwargs["days"] = days
    
    while idx < len(s):
        
        idx,pfx,v = _parse_duration_number_part(s,idx)
                
        if pfx == 'H':
            kwargs["hours"] = v
            
        elif pfx == 'M':
            kwargs["minutes"] = v
            
        elif pfx == 'S':
            kwargs["seconds"] = v
        
        else:
            if idx != len(s):
                raise SyntaxError("ISO 8601 duration [%s] contains invalid date part [%s]."%(s,pfx))
    
    return timedelta(**kwargs)

def format_duration(dt):
    '''
    Format a timedelta instance as an ISO-8601 duration.
    
    :param dt: A positive timedelta instance
    '''
    
    if dt.days < 0 or dt.seconds < 0 or dt.microseconds < 0:
        raise ValueError("Cannot format negative timedelta [%s] as ISO-8601 duration."%dt)
    
    if dt.days != 0:
        if dt.seconds == 0 and dt.microseconds == 0:
            return "P%dD" % dt.days
        else:
            ret = "P%dDT" % dt.days
    else:   
        ret = 'PT'

    if dt.seconds != 0:
        
        hours = (dt.seconds // 3600) % 24

        if hours !=0:
            ret += "%dH" % hours

        mins = (dt.seconds // 60) % 60

        if mins !=0:
            ret += "%dM" % mins
        
        secs = dt.seconds % 60
    
    else:
        secs = 0
    
    
    if dt.microseconds == 0:
        if secs != 0 or ret == 'PT':
            ret += "%dS" % secs
    elif dt.microseconds % 1000 == 0:
        ret += "%d.%03dS" % (secs,dt.microseconds//1000)
    else:
        ret += "%d.%06dS" % (secs,dt.microseconds)

    return ret

def get_os_timezone():
    '''
    Inspect the system timezone, wither by inspecting the symbolic link
    ``/etc/localtime`` (RedHat style) or reading the file ``/etc/timezone`` (Debian style)
    
    If both methods fail, ``UTC`` is returned.
    
    :return: An Olson timezone name like ``Europe/Berlin``.
    '''
    
    ret = None
    
    try:
        if stat.S_ISLNK(os.lstat('/etc/localtime').st_mode):
            tzpath = os.readlink("/etc/localtime")
            if tzpath.startswith("/usr/share/zoneinfo/"):
                ret = tzpath[20:]
    except:
        log.exception("Error inspecting [/etc/localtime], will try [/etc/timezone].")
    
    if ret is None:
        try:
            with io.open("/etc/timezone") as fp:
                ret = fp.read().strip()
        except:
            log.exception("Error inspecting [/etc/timezone], local timezone could not be detected, returning [UTC].")

    if ret is None:
        return "UTC"
    else:
        return ret