'''
..
  $Id$
  
  Basic python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import os
import re

def read_file(filename:str,binary:bool = False, encoding = "utf-8"):
    '''
    Read the contents of a whole file either as bytes or as a string.
    
    :param filename: The name of the file to read.
    :param binary: If true, read the file in binary mode and return a bytes object.
    :param encoding: The encoding for non-binary processing, which defaults to UTF-8.
    '''
    with open(filename,"rb" if binary else "r",encoding=None if binary else encoding) as fh:
        return fh.read()

def get_deb_version(package:str,changelog_filename:str):
    '''
    Read the debian version and debian release tag from a changelog file
    
    :param package: The name of the source package
    :param changelog_filename: The path to the debian changelog file
    '''
    with open(changelog_filename) as fd:
        line = fd.readline() 
        
        pattern = '^'+package+'\\s*\\(([^-]*)-([^)]*)\\).*\n'
        #print("Matching first line of changelog against pattern [%s]" % pattern)
        res = re.match(pattern,line)
                       
        if not res:
            raise SyntaxError("debian changelog line [%s] is malformatted and/or package name not [%s]" % (line[:-1], package) )
                
        return res.group(1),res.group(2)

def get_rpm_version(package:str,spec_filename:str):
    '''
    Reads the spec file and parses the first line with Version: resp. Release: 
    
    :param package: The name of the source package
    :param spec_filename: The path to the rpm spec file
    '''
    with open(spec_filename) as fd:
        version = ""
        rel = ""

        for line in fd:
            if len(version) == 0 and line.startswith("Version: "):
                version = line.split()[1]
            if len(rel) == 0 and line.startswith("Release: "):
                rel = line.split()[1]
        
        if len(version) == 0:
            raise SyntaxError("File file [%s] did not provide a version" % spec_filename)
            
        return version,rel
    
def get_deb_or_rpm_version(package:str,changelog_filename:str,spec_filename:str):
    '''
    If changelog_filename exists, calls get_deb_version
    Otherwise, if spec_filename exists calls get_rpm_version
    Otherwise throws up 
    
    :param package: The name of the source package
    :param changelog_filename: The path to the debian changelog file
    :param spec_filename: The path to the rpm spec file
    '''
    if os.path.exists(changelog_filename):
        return get_deb_version(package, changelog_filename)
    
    if os.path.exists(spec_filename):
        return get_rpm_version(package, spec_filename)

    raise SyntaxError("Neither changelog [%s] nor spec file [%s] exist" % (changelog_filename, spec_filename) )
