'''
..
  $Id$
  
  Basic python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 21.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''
import logging

log = logging.getLogger(__name__)

def str2bool(v):
    '''
    Parse a string to a boolean value.
    This function exists, because ``bool("false")`` evaluate to
    ``True`` in python.
    
    :param v: A string.
    :return: True, if the passed string is case insensitively equal to
           ``t``, ``1``, ``true`` or ``yes``.
    '''
    return v.lower() in ("yes", "true", "t", "1")

def get_with_type(dictionary,key,ttype):
    '''
    Get a key from a dictionary and try to convert the value to the given
    type, if the value is a string.
    
    :param dictionary: A dictionary.
    :param key: The key to look up.
    :param ttype: The type to convert the value to, if it is a string.
    :return: The value, which is of the given type or None, if the key 
    does not exist in ``dictionary``
    '''

    ret = dictionary.get(key)
    
    if ret is not None:
        try:
            if isinstance(ret,ttype):
                return ret
            elif ttype == bool:
                return str2bool(ret)
            else:
                return ttype(ret)
        except:
            log.warn("Value [%s] is not of type [%s], returning None."%(ret,ttype))

    return None


def get_or_default(dictionary,key,defval):
    '''
    Get a key from a dictionary and try to convert the value to the
    type of the given default value, if the value is a string.
    
    If the key does not exist in the dictionary, return the default value.
    
    :param dictionary: A dictionary.
    :param key: The key to look up.
    :param defval: The default value.
    :return: The value, which is of the given type or None, if the key
             does not exist in ``dictionary``
    '''
    
    ttype = type(defval)
    ret = dictionary.get(key)
    
    if ret is not None:
        try:
            if isinstance(ret,ttype):
                return ret
            elif ttype == bool:
                return str2bool(ret)
            else:
                return ttype(ret)
        except:
            log.warn("Config value [%s] is not of type [%s], using default value [%d]."%(ret,ttype,defval))

    return defval

def extension(clazz):
    '''
    Extend a class by a given method. This is used to replace the infamous
    visitor pattern in JAVA.

    It is used like::
    
      @extension(Dog)
      def sound(self):
          print("Wooff")
          
      @extension(Cat)
      def sound(self):
          print("Miaou")
      
    This creates sound methods on the Cat and Dog classes.
        
    :param clazz: The class to be extended
    :return: A decorator function which, when given a function object, will add that 
             function object to ``clazz``.
    '''
    def inner_decorator(func):
        setattr(clazz,func.__name__,func)
        return func
    return inner_decorator
