'''
..
  $Id$
  
  Basic python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''
import io
import json
import logging.handlers
import logging.config
import os
import sys

def configureRootLogger(handler: logging.Handler, logLevel: int):
    '''
    Configure the root logger with a given log handler and a log level.
    
    :param handler: The handler to use for the root logger. Be aware, that the root
    handler must have a least the given logeLvel configured, otherwise
    you won't receive any output.
    :param logLevel: The log level, which may be ``logging.INFO``, ``loggging.DEBUG``, ...
    '''
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(logLevel)

def configureAuxilaryLoggers(loggerNames, logLevel: int):
    '''
    Configure the loggers for all given names to the given log level.
    If called at the beginning of your program, the logger will inherit the handler
    from the root logger.
    
    :param loggerNames: A list or iterable containing all logger to be configured.
    :param logLevel: The log level, which may be ``logging.INFO``, ``loggging.DEBUG``, ...
    '''
    for loggerName in loggerNames:
        logger = logging.getLogger(loggerName)
        logger.setLevel(logLevel)

def configureSyslog(ident:str, logLevel:int = logging.INFO, facility=logging.handlers.SysLogHandler.LOG_DAEMON, debugNames=None):
    '''
    Configure the logging system to syslog with an optional list of debug log names.
    
    :param ident: The program ident used as a prefix for all log lines.
    :param logLevel: The root log level, which may be ``logging.INFO``, ``loggging.DEBUG``, ...
    :param facility: The syslog facility, for the most part, the default value ``LOG_DAEMON`` is OK.
    :param debugNames: A list or iterable containing all logger to be configured to ``logging.DEBUG``.
                       If given, the log handler will at least be configured to log level ``logging.DEBUG``.
    '''
    handler = logging.handlers.SysLogHandler(facility=facility,address='/dev/log')
    formatter = logging.Formatter(ident+' %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    
    if debugNames and logLevel > logging.DEBUG:
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logLevel)

    configureRootLogger(handler,logLevel)
    
    if debugNames:
        configureAuxilaryLoggers(debugNames,logging.DEBUG)

def configureStderr(logLevel: int = logging.INFO, debugNames=None):
    '''
    Configure the logging system to stderr with an optional list of debug log names.
    
    :param logLevel: The root log level, which may be ``logging.INFO``, ``loggging.DEBUG``, ...
    :param facility: The syslog facility, for the most part, the default value ``LOG_DAEMON`` is OK.
    :param debugNames: A list or iterable containing all logger to be configured to ``logging.DEBUG``.
                       If given, the log handler will at least be configured to log level ``logging.DEBUG``.
    '''
    handler = logging.StreamHandler(sys.stderr)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s') 
    handler.setFormatter(formatter)
    
    if debugNames and logLevel > logging.DEBUG:
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logLevel)
    
    configureRootLogger(handler,logLevel)
    
    if debugNames:
        configureAuxilaryLoggers(debugNames,logging.DEBUG)      

def configureTestLogging(file):
    '''
    Load a logging configuration from the file ``logging.json`` in the same
    directory as the passed in file name.
    
    :param file: The file name of the test module.
    '''
    configureLogging(os.path.join(os.path.dirname(os.path.abspath(file)),"logging.json"))


def configureLogging(jsonPath):
    '''
    Load a JSON file and configure the python std library logging using the
    dictionary contained in the given JSON file.
    
    :param jsonPath: The path of the JSON file to read in.
    '''
    with io.open(jsonPath,mode='r') as inp:
        logging.config.dictConfig(json.load(inp))
